# -*- coding: utf-8 -*-
#
# Copyright (C) 2021 CERN.
# Copyright (C) 2021-2026 Northwestern University.
# Copyright (C) 2021 TU Wien.
#
# Invenio-Requests is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Requests service configuration."""

from invenio_records_resources.services import (
    RecordServiceConfig,
    SearchOptions,
    pagination_endpoint_links,
)
from invenio_records_resources.services.base.config import (
    ConfiguratorMixin,
    FromConfig,
    FromConfigSearchOptions,
    SearchOptionsMixin,
)

from invenio_requests.services.requests import facets

from ...customizations import RequestActions
from ...records.api import Request
from ..links import (
    ActionsEndpointLinks,
    RequestEndpointLink,
    RequestListOfCommentsEndpointLink,
    RequestTypeDependentEndpointLink,
)
from ..permissions import PermissionPolicy
from .components import (
    EntityReferencesComponent,
    RequestDataComponent,
    RequestLockComponent,
    RequestNumberComponent,
    RequestPayloadComponent,
    RequestReviewersComponent,
)
from .params import IsOpenParam, ReferenceFilterParam, SharedOrMyRequestsParam
from .results import RequestItem, RequestList


def _is_action_available(request, context):
    """Check if the given action is available on the request."""
    action = context.get("action")
    identity = context.get("identity")
    permission_policy_cls = context.get("permission_policy_cls")
    permission = permission_policy_cls(f"action_{action}", request=request)
    return RequestActions.can_execute(request, action) and permission.allows(identity)


class RequestSearchOptions(SearchOptions, SearchOptionsMixin):
    """Search options."""

    params_interpreters_cls = SearchOptions.params_interpreters_cls + [
        ReferenceFilterParam.factory(param="created_by", field="created_by"),
        ReferenceFilterParam.factory(param="receiver", field="receiver"),
        ReferenceFilterParam.factory(param="topic", field="topic"),
        IsOpenParam.factory("is_open"),
    ]

    facets = {
        "type": facets.type,
        "status": facets.status,
    }


class UserRequestSearchOptions(RequestSearchOptions):
    """User request search options."""

    params_interpreters_cls = RequestSearchOptions.params_interpreters_cls + [
        SharedOrMyRequestsParam,
    ]


class RequestsServiceConfig(RecordServiceConfig, ConfiguratorMixin):
    """Requests service configuration."""

    service_id = "requests"

    # common configuration
    permission_policy_cls = FromConfig(
        "REQUESTS_PERMISSION_POLICY", default=PermissionPolicy
    )
    result_item_cls = RequestItem
    result_list_cls = RequestList
    search = FromConfigSearchOptions(
        config_key="REQUESTS_SEARCH",
        sort_key="REQUESTS_SORT_OPTIONS",
        facet_key="REQUESTS_FACETS",
        search_option_cls=RequestSearchOptions,
    )

    # user requests search configuration
    search_user_requests = FromConfigSearchOptions(
        config_key="REQUESTS_SEARCH",
        sort_key="REQUESTS_SORT_OPTIONS",
        facet_key="REQUESTS_FACETS",
        search_option_cls=UserRequestSearchOptions,
    )

    # request-specific configuration
    record_cls = Request  # needed for model queries
    schema = None  # stored in the API classes, for customization
    indexer_queue_name = "requests"
    index_dumper = None

    # links configuration
    links_item = {
        "self": RequestEndpointLink("requests.read"),
        "self_html": RequestTypeDependentEndpointLink(
            key="self_html",
            request_retriever=lambda obj, vars: obj,
            request_type_retriever=lambda obj, vars: obj.type,
        ),
        # Note that `request_events` is the name of the blueprint for
        # the RequestCommentsResource actually.
        "comments": RequestListOfCommentsEndpointLink("request_events.create"),
        "timeline": RequestListOfCommentsEndpointLink("request_events.search"),
        "timeline_focused": RequestListOfCommentsEndpointLink(
            "request_events.focused_list"
        ),
        "lock": RequestEndpointLink("requests.lock_request"),
        "unlock": RequestEndpointLink("requests.unlock_request"),
        "actions": ActionsEndpointLinks(
            RequestEndpointLink(
                "requests.execute_action",
                # "id" would have been added by RequestEndpointLink but
                # it's more explicit this way
                params=["id", "action"],
                when=_is_action_available,
            )
        ),
    }
    links_search = pagination_endpoint_links("requests.search")
    links_user_requests_search = pagination_endpoint_links(
        "requests.search_user_requests"
    )

    payload_schema_cls = None

    # TODO: discuss conflict between this and custom request service.
    #  Does it create issues?
    components = FromConfig(
        "REQUESTS_SERVICE_COMPONENTS",
        default=[
            # Order of components is important!
            RequestPayloadComponent,
            RequestDataComponent,
            RequestReviewersComponent,
            EntityReferencesComponent,
            RequestNumberComponent,
            RequestLockComponent,
        ],
    )



# -*- coding: utf-8 -*-
#
# Copyright (C) 2022 CERN.
# Copyright (C) 2023 Graz University of Technology.
#
# Invenio-Requests is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Request views module."""

from flask import Blueprint

from .api import create_request_events_bp, create_request_files_bp, create_requests_bp
from .ui import create_ui_blueprint

blueprint = Blueprint("invenio-requests-ext", __name__)


__all__ = (
    "blueprint",
    "create_ui_blueprint",
    "create_requests_bp",
    "create_request_events_bp",
    "create_request_files_bp",
)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2025 CERN.
#
# Invenio-Requests is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""RequestFiles Resource."""

from flask import g
from flask_resources import resource_requestctx, response_handler, route
from invenio_records_resources.resources import RecordResource
from invenio_records_resources.resources.files.resource import request_stream
from invenio_records_resources.resources.records.resource import (
    request_extra_args,
    request_headers,
    request_view_args,
)


#
# Resource
#
class RequestFilesResource(RecordResource):
    """Resource for Request files."""

    def create_url_rules(self):
        """Create the URL rules for the record resource."""
        routes = self.config.routes
        return [
            route("PUT", routes["create"], self.create),
            route("DELETE", routes["item"], self.delete),
            route("GET", routes["item_content"], self.read),
        ]

    @request_extra_args
    @request_headers
    @request_view_args
    @request_stream
    @response_handler()
    def create(self):
        """Create a file."""
        item = self.service.create_file(
            identity=g.identity,
            id_=resource_requestctx.view_args["id"],
            key=resource_requestctx.view_args["key"],
            stream=resource_requestctx.data["request_stream"],
            content_length=resource_requestctx.data["request_content_length"],
        )
        return item.to_dict(), 200

    @request_view_args
    def delete(self):
        """Delete a file."""
        self.service.delete_file(
            identity=g.identity,
            id_=resource_requestctx.view_args["id"],
            file_key=resource_requestctx.view_args["key"],
        )
        return "", 204

    @request_view_args
    def read(self):
        """Read a file."""
        file = self.service.read_file(
            identity=g.identity,
            id_=resource_requestctx.view_args["id"],
            file_key=resource_requestctx.view_args["key"],
        )
        return file.send_file(as_attachment=True, restricted=True), 200



# -*- coding: utf-8 -*-
#
# Copyright (C) 2025 CERN.
#
# Invenio-Requests is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""RequestFiles Resource Configuration."""

from flask_resources import HTTPJSONException, create_error_handler
from invenio_records_resources.resources import RecordResourceConfig
from marshmallow import fields

from invenio_requests.services.files.errors import (
    RequestFileNotFoundError,
    RequestFileSizeLimitError,
)


class RequestFilesResourceConfig(RecordResourceConfig):
    """Request Files resource configuration."""

    blueprint_name = "request_files"
    url_prefix = "/requests"
    routes = {
        "create": "/<id>/files/upload/<key>",
        "item": "/<id>/files/<key>",
        "item_content": "/<id>/files/<key>/content",
    }

    request_view_args = {
        "id": fields.UUID(),
        "key": fields.Str(),
    }

    response_handlers = {
        "application/vnd.inveniordm.v1+json": RecordResourceConfig.response_handlers[
            "application/json"
        ],
        **RecordResourceConfig.response_handlers,
    }

    error_handlers = {
        **RecordResourceConfig.error_handlers,
        RequestFileSizeLimitError: create_error_handler(
            lambda e: HTTPJSONException(
                code=400,
                description=str(e),
            )
        ),
        RequestFileNotFoundError: create_error_handler(
            lambda e: HTTPJSONException(
                code=404,
                description=str(e),
            )
        ),
    }



# -*- coding: utf-8 -*-
#
# Copyright (C) 2025 CERN.
#
# Invenio-Requests is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Requests service configuration."""

from invenio_records_resources.services import (
    FileServiceConfig,
    Link,
)
from invenio_records_resources.services.base.config import ConfiguratorMixin, FromConfig
from invenio_records_resources.services.files.links import FileEndpointLink
from invenio_records_resources.services.records.results import RecordItem

from ...records.api import Request, RequestFile
from ..permissions import PermissionPolicy
from ..schemas import RequestFileSchema


class RequestFileItem(RecordItem):
    """RequestFile result item."""

    @property
    def id(self):
        """Id property."""
        return self._record.id


class RequestFileLink(Link):
    """Link variables setter for RequestFile links."""

    @staticmethod
    def vars(obj, vars):
        """Variables for the URI template."""
        vars.update({"request_id": obj.model.record_id, "key": obj.key})
        # Remark: we do not call `request_type._update_link_config`
        # because we do not want file links to be modified depending on the context (e.g. `/me`)


class RequestFilesServiceConfig(FileServiceConfig, ConfiguratorMixin):
    """Requests Files service configuration."""

    service_id = "request_files"

    # common configuration
    permission_action_prefix = ""
    permission_policy_cls = FromConfig(
        "REQUESTS_PERMISSION_POLICY", default=PermissionPolicy
    )
    result_item_cls = RequestFileItem

    # request files-specific configuration
    record_cls = RequestFile  # needed for model queries
    schema = RequestFileSchema  # stored in the API classes, for customization
    request_cls = Request
    indexer_queue_name = "files"

    # links configuration / ResultItem configurations
    links_item = {
        "self": RequestFileLink("{+api}/requests/{request_id}/files/{key}"),
        "content": RequestFileLink("{+api}/requests/{request_id}/files/{key}/content"),
        "download_html": RequestFileLink("{+ui}/requests/{request_id}/files/{key}"),
    }

    file_links_item = {
        "self": FileEndpointLink("request_files.read", params=["pid_value", "key"]),
    }

    components = FromConfig(
        "REQUESTS_FILES_SERVICE_COMPONENTS",
        default=[],
    )



#
# Copyright (C) 2025 CERN.
#
# Invenio-Requests is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Create request files table."""

import sqlalchemy as sa
import sqlalchemy_utils
from alembic import op
from sqlalchemy.dialects import mysql, postgresql

# revision identifiers, used by Alembic.
revision = "1763728177"
down_revision = "1759321170"
branch_labels = ()
depends_on = "8ae99b034410"


def upgrade():
    """Upgrade database."""
    op.create_table(
        "request_files",
        sa.Column("id", sqlalchemy_utils.types.uuid.UUIDType(), nullable=False),
        sa.Column(
            "json",
            sa.JSON()
            .with_variant(sqlalchemy_utils.types.json.JSONType(), "mysql")
            .with_variant(
                postgresql.JSONB(none_as_null=True, astext_type=sa.Text()), "postgresql"
            )
            .with_variant(sqlalchemy_utils.types.json.JSONType(), "sqlite"),
            nullable=True,
        ),
        sa.Column("version_id", sa.Integer(), nullable=False),
        sa.Column(
            "created",
            sa.DateTime().with_variant(mysql.DATETIME(fsp=6), "mysql"),
            nullable=False,
        ),
        sa.Column(
            "updated",
            sa.DateTime().with_variant(mysql.DATETIME(fsp=6), "mysql"),
            nullable=False,
        ),
        sa.Column(
            "key",
            sa.Text().with_variant(mysql.VARCHAR(length=255), "mysql"),
            nullable=False,
        ),
        sa.Column("record_id", sqlalchemy_utils.types.uuid.UUIDType(), nullable=False),
        sa.Column(
            "object_version_id", sqlalchemy_utils.types.uuid.UUIDType(), nullable=True
        ),
        sa.ForeignKeyConstraint(
            ["object_version_id"],
            ["files_object.version_id"],
            name=op.f("fk_request_files_object_version_id_files_object"),
            ondelete="RESTRICT",
        ),
        sa.ForeignKeyConstraint(
            ["record_id"],
            ["request_metadata.id"],
            name=op.f("fk_request_files_record_id_request_metadata"),
            ondelete="RESTRICT",
        ),
        sa.PrimaryKeyConstraint("id", name=op.f("pk_request_files")),
    )
    op.create_index(
        op.f("ix_request_files_object_version_id"),
        "request_files",
        ["object_version_id"],
        unique=False,
    )
    op.create_index(
        op.f("ix_request_files_record_id"), "request_files", ["record_id"], unique=False
    )
    op.create_index(
        "uidx_request_files_record_id_key",
        "request_files",
        ["record_id", "key"],
        unique=True,
    )
    op.add_column(
        "request_metadata",
        sa.Column("bucket_id", sqlalchemy_utils.types.uuid.UUIDType(), nullable=True),
    )
    op.create_index(
        op.f("ix_request_metadata_bucket_id"),
        "request_metadata",
        ["bucket_id"],
        unique=False,
    )
    op.create_foreign_key(
        op.f("fk_request_metadata_bucket_id_files_bucket"),
        "request_metadata",
        "files_bucket",
        ["bucket_id"],
        ["id"],
        ondelete="RESTRICT",
    )


def downgrade():
    """Downgrade database."""
    op.drop_index(op.f("ix_request_metadata_bucket_id"), table_name="request_metadata")
    op.drop_constraint(
        op.f("fk_request_metadata_bucket_id_files_bucket"),
        "request_metadata",
        type_="foreignkey",
    )
    op.drop_column("request_metadata", "bucket_id")
    op.drop_index("uidx_request_files_record_id_key", table_name="request_files")
    op.drop_index(op.f("ix_request_files_record_id"), table_name="request_files")
    op.drop_index(
        op.f("ix_request_files_object_version_id"), table_name="request_files"
    )
    op.drop_table("request_files")



# -*- coding: utf-8 -*-
#
# Copyright (C) 2025 CERN.
#
# Invenio-Requests is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Request Files views module."""

from .ui import create_ui_blueprint

__all__ = ("create_ui_blueprint",)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021 - 2022 TU Wien.
#
# Invenio-Requests is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Module for entity resolvers."""

# NOTE: no imports made available here because that introduces circular imports
__all__ = ()



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021-2025 CERN.
# Copyright (C) 2021 - 2022 TU Wien.
# Copyright (C) 2025 KTH Royal Institute of Technology.
#
# Invenio-Requests is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Invenio module for generic and customizable requests."""

from invenio_i18n import lazy_gettext as _
from invenio_users_resources.entity_resolvers import GroupResolver, UserResolver

from invenio_requests.services.requests import facets
from invenio_requests.services.requests.components import (
    RequestCommentFileCleanupComponent,
    RequestCommentFileValidationComponent,
)

from .customizations import CommentEventType, LogEventType, ReviewersUpdatedType
from .services.permissions import PermissionPolicy

REQUESTS_PERMISSION_POLICY = PermissionPolicy
"""Override the default requests/comments permission policy."""

REQUESTS_REGISTERED_TYPES = []
"""Configuration for registered Request Types."""

REQUESTS_REGISTERED_EVENT_TYPES = [
    LogEventType(),
    CommentEventType(),
    ReviewersUpdatedType(),
]
"""Configuration for registered Request Event Types."""

REQUESTS_ENTITY_RESOLVERS = [UserResolver(), GroupResolver()]
"""Registered resolvers for resolving/creating references in request metadata."""

REQUESTS_ROUTES = {
    "download_file_html": "/requests/<uuid:pid_value>/files/<path:file_key>",
}
"""Invenio requests ui endpoints."""

REQUESTS_SEARCH = {
    "facets": ["type", "status"],
    "sort": ["bestmatch", "newest", "oldest", "newestactivity", "oldestactivity"],
}
"""Requests search default configuration."""

REQUESTS_SORT_OPTIONS = {
    "bestmatch": dict(
        title=_("Best match"),
        fields=["_score"],  # search defaults to desc on `_score` field
    ),
    "newest": dict(
        title=_("Newest"),
        fields=["-created"],
    ),
    "oldest": dict(
        title=_("Oldest"),
        fields=["created"],
    ),
    "newestactivity": dict(
        title=_("Newest activity"),
        fields=["-last_activity_at"],
    ),
    "oldestactivity": dict(
        title=_("Oldest activity"),
        fields=["last_activity_at"],
    ),
}
"""Definitions of available request sort options."""

REQUESTS_FACETS = {
    "type": {
        "facet": facets.type,
        "ui": {
            "field": "type",
        },
    },
    "status": {
        "facet": facets.status,
        "ui": {
            "field": "status",
        },
    },
}
"""Invenio requests facets."""

REQUESTS_TIMELINE_PAGE_SIZE = 10
"""Amount of items per page on the request details timeline"""


REQUESTS_MODERATION_ROLE = "administration-moderation"
"""ID of the Role used for moderation."""


#
# User moderation administration
#
REQUESTS_USER_MODERATION_SEARCH = {
    "facets": ["status", "is_open"],
    "sort": ["bestmatch", "newest", "oldest"],
}
"""Community requests search configuration (i.e list of community requests)"""

REQUESTS_USER_MODERATION_SORT_OPTIONS = {
    "bestmatch": dict(
        title=_("Best match"),
        fields=["_score"],  # ES defaults to desc on `_score` field
    ),
    "newest": dict(
        title=_("Newest"),
        fields=["-created"],
    ),
    "oldest": dict(
        title=_("Oldest"),
        fields=["created"],
    ),
}
"""Definitions of available record sort options."""

REQUESTS_USER_MODERATION_FACETS = {
    "status": {
        "facet": facets.status,
        "ui": {
            "field": "status",
        },
    },
    "is_open": {"facet": facets.is_open, "ui": {"field": "is_open"}},
}
"""Available facets defined for this module."""

REQUESTS_REVIEWERS_ENABLED = False
"""Enable reviewers for requests."""

REQUESTS_REVIEWERS_MAX_NUMBER = 15
"""Maximum number of reviewers allowed for a request."""

REQUESTS_LOCKING_ENABLED = False
"""Enable locking/unlocking for request conversations."""

REQUESTS_COMMENT_PREVIEW_LIMIT = 5
"""Number of most recent child comments to inline in parent's search index.

This limits the size of indexed documents when comments have many replies.
Additional replies can be loaded via pagination.
"""

REQUESTS_FILES_DEFAULT_QUOTA_SIZE = 100 * 10**6  # 100MB
REQUESTS_FILES_DEFAULT_MAX_FILE_SIZE = 10 * 10**6  # 10MB

REQUESTS_EVENTS_SERVICE_COMPONENTS = [
    RequestCommentFileValidationComponent,
    RequestCommentFileCleanupComponent,
]

REQUESTS_COMMENTS_ALLOWED_EXTRA_HTML_TAGS = ["img"]
"""Extend allowed HTML tags list for requests comments content."""

REQUESTS_COMMENTS_ALLOWED_EXTRA_HTML_ATTRS = {
    "img": ["src", "alt", "width", "height"],
}
"""Extend allowed HTML attrs list for requests comments content."""



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021 TU Wien.
#
# Invenio-Requests is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Search dumpers, for transforming to and from versions to index."""

from .calculated import CalculatedFieldDumperExt
from .granttokens import GrantTokensDumperExt
from .parentchild import ParentChildDumperExt

__all__ = (
    "CalculatedFieldDumperExt",
    "GrantTokensDumperExt",
    "ParentChildDumperExt",
)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021 CERN.
# Copyright (C) 2021 TU Wien.
#
# Invenio-Requests is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Request comment services module."""

from .config import RequestEventsServiceConfig
from .service import RequestEventsService

__all__ = (
    "RequestEventsService",
    "RequestEventsServiceConfig",
)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2023 Graz University of Technology.
#
# Invenio-Requests is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Recipient filters for notifications."""

from invenio_notifications.services.filters import RecipientFilter
from invenio_records.dictutils import dict_lookup


class UserRecipientFilter(RecipientFilter):
    """Recipient filter based on user."""

    def __init__(self, key):
        """Initialize with key for user lookup."""
        super().__init__()
        self._key = key

    def __call__(self, notification, recipients):
        """Filter recipients."""
        user = dict_lookup(notification.context, self._key)
        if not isinstance(user, dict):
            # e.g. resolved email entity
            return recipients

        # lookup (non) expanded user field
        user_id = user.get("user") or user.get("id")

        if user_id in recipients:
            del recipients[user_id]

        return recipients



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021 CERN.
# Copyright (C) 2021 TU Wien.
# Copyright (C) 2025 Northwestern University.
#
# Invenio-Requests is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Services module."""

from .components import RequestNumberComponent
from .config import RequestsServiceConfig
from .results import RequestItem, RequestList
from .service import RequestsService

__all__ = (
    "RequestNumberComponent",
    "RequestItem",
    "RequestList",
    "RequestsService",
    "RequestsServiceConfig",
)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021 TU Wien.
#
# Invenio-Requests is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Custom exceptions used in the Invenio-Requests module."""

from invenio_i18n import lazy_gettext as _
from invenio_records_resources.services.errors import PermissionDeniedError


class ActionError(Exception):
    """Exception indicating an error related to the action."""

    def __init__(self, action, reason):
        """Constructor.

        :param action: The name of the action in question.
        :param reason: Description of what went wrong.
        """
        self.action = action
        self.reason = reason

    def __str__(self):
        """Return str(self)."""
        return f"Error for action '{self.action}': {self.reason}"


class NoSuchActionError(ActionError):
    """There was no such action available."""

    def __init__(self, action, reason=None):
        """Constructor.

        :param action: The name of the action.
        """
        reason = reason or "No such action available"
        super().__init__(action, reason)


class CannotExecuteActionError(ActionError):
    """The selected action could not be executed."""

    def __init__(self, action, reason=None):
        """Constructor.

        :param action: The name of the action.
        """
        reason = reason or "Could not execute the action"
        super().__init__(action, reason)


class RequestLockedError(PermissionDeniedError):
    """Exception indicating that the request is locked."""

    def __init__(self, description=None):
        """Constructor.

        :param message: The message to display.
        """
        self._description = description or _("The request is locked.")

    @property
    def description(self):
        """Return the description."""
        return self._description

    def __str__(self):
        """Return str(self)."""
        return self.description


class RequestEventPermissionError(PermissionDeniedError):
    """Exception indicating that the request event permission is denied."""

    def __init__(self, description=None):
        """Constructor.

        :param message: The message to display.
        """
        self._description = description or _("Permission denied.")

    @property
    def description(self):
        """Return the description."""
        return self._description

    def __str__(self):
        """Return str(self)."""
        return self.description


class ChildrenNotSupportedError(Exception):
    """Exception raised when children are attempted on an event type that doesn't support it."""

    def __init__(self, event_type, message=None):
        """Constructor.

        :param event_type: The event type that doesn't support children.
        :param message: Optional custom error message.
        """
        self.event_type = event_type
        self.message = message or _(
            (f"Event type '{event_type}' does not support event replies.")
        )
        super().__init__(self.message)


class NestedChildrenNotAllowedError(Exception):
    """Exception raised when attempting to create nested children (reply to a reply)."""

    def __init__(self, message=None):
        """Constructor.

        :param message: Optional custom error message.
        """
        self.message = message or _(
            (
                "Nested children are not allowed. "
                "You cannot reply to a comment that is already a reply. "
                "Only one level of parent-child relationships is supported."
            )
        )
        super().__init__(self.message)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2026 CERN.
#
# Invenio-Requests is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Files field allow for explicit setting after initialization."""

from invenio_records_resources.records.systemfields.files.field import FilesField


class RequestFilesField(FilesField):
    """Request Files system field."""

    def __init__(self, *args, **kwargs):
        """Initialize the RequestFilesField."""
        super().__init__(*args, **kwargs)

    #
    # Data descriptor methods (i.e. attribute access)
    #
    def __set__(self, record, value):
        """Set the request's files field."""
        assert record is not None

        self.set_dictkey(record, value)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021 CERN.
# Copyright (C) 2021 TU Wien.
#
# Invenio-Requests is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Component for creating request numbers."""

from uuid import UUID

from flask import current_app
from invenio_i18n import _
from invenio_records_resources.services.errors import ValidationErrorGroup
from invenio_records_resources.services.records.components import (
    DataComponent,
    ServiceComponent,
)
from marshmallow import ValidationError

from invenio_requests.customizations.event_types import (
    LogEventType,
    ReviewersUpdatedType,
)
from invenio_requests.proxies import (
    current_events_service,
    current_request_files_service,
)
from invenio_requests.records.api import RequestFile


class RequestNumberComponent(ServiceComponent):
    """Component for assigning request numbers to new requests."""

    def create(self, identity, data=None, record=None, **kwargs):
        """Create identifier when record is created."""
        type(record).number.assign(record)


class EntityReferencesComponent(ServiceComponent):
    """Component for initializing a request's entity references."""

    def create(self, identity, data=None, record=None, **kwargs):
        """Initialize the entity reference fields of a request."""
        for field in ("created_by", "receiver", "topic"):
            if field in kwargs:
                setattr(record, field, kwargs[field])


class RequestDataComponent(DataComponent):
    """Request variant of DataComponent using dynamic schema."""

    def update(self, identity, data=None, record=None, **kwargs):
        """Update an existing record (request)."""
        if record.status == "created":
            keys = ("title", "description", "payload", "receiver", "topic")
        else:
            keys = ("title", "description")
        if current_app.config["REQUESTS_LOCKING_ENABLED"]:
            keys = keys + ("is_locked",)

        for k in keys:
            if k in data:
                record[k] = data[k]


class RequestReviewersComponent(ServiceComponent):
    """Component for handling request reviewers."""

    def _reviewers_updated(self, previous_reviewers, new_reviewers):
        """Determine reviewers change type: added, removed, updated, unchanged."""

        def _normalize(reviewers):
            """Convert reviewers into a set of string identifiers."""
            normalized = set()
            for r in reviewers:
                if "user" in r:
                    normalized.add(f"user:{r['user']}")
                elif "group" in r:
                    normalized.add(f"group:{r['group']}")
            return normalized

        prev_set = _normalize(previous_reviewers)
        new_set = _normalize(new_reviewers)

        added = new_set - prev_set
        removed = prev_set - new_set

        if added and not removed:
            return "added", [
                r
                for r in new_reviewers
                if (
                    ("user" in r and f"user:{r['user']}" in added)
                    or ("group" in r and f"group:{r['group']}" in added)
                )
            ]
        elif removed and not added:
            return "removed", [
                r
                for r in previous_reviewers
                if (
                    ("user" in r and f"user:{r['user']}" in removed)
                    or ("group" in r and f"group:{r['group']}" in removed)
                )
            ]
        elif added and removed:
            return "updated", list(new_reviewers)
        else:
            return "unchanged", list(new_reviewers)

    def _validate_reviewers(self, reviewers):
        """Validate the reviewers data."""
        reviewers_enabled = current_app.config["REQUESTS_REVIEWERS_ENABLED"]
        reviewers_groups_enabled = current_app.config["USERS_RESOURCES_GROUPS_ENABLED"]
        max_reviewers = current_app.config["REQUESTS_REVIEWERS_MAX_NUMBER"]

        if not reviewers_enabled:
            raise ValidationError(_("Reviewers are not enabled for this request type."))
        if not reviewers_groups_enabled:
            for reviewer in reviewers:
                if "group" in reviewer:
                    raise ValidationError(_("Group reviewers are not enabled."))

        if len(reviewers) > max_reviewers:
            raise ValidationError(
                _(f"You can only add up to {max_reviewers} reviewers.")
            )

    def _ensure_no_duplicates(self, reviewers):
        """Ensure there are no duplicate reviewers.

        The code is preserving the original order of reviewers.
        """
        seen = set()
        unique_objs = []
        for d in reviewers:
            t = tuple(sorted(d.items()))
            if t not in seen:
                seen.add(t)
                unique_objs.append(d)
        return unique_objs

    def update(self, identity, data=None, record=None, uow=None, **kwargs):
        """Update the reviewers of a request."""
        if "reviewers" in data:
            # ensure there are not duplicates
            new_reviewers = self._ensure_no_duplicates(data["reviewers"])
            self._validate_reviewers(new_reviewers)
            self.service.require_permission(identity, f"action_accept", request=record)

            event_type, updated_reviewers = self._reviewers_updated(
                record.get("reviewers", []), new_reviewers
            )
            if not event_type == "unchanged":
                event = ReviewersUpdatedType(
                    payload=dict(
                        event="reviewers_updated",
                        content=_(f"{event_type} a reviewer"),
                        reviewers=updated_reviewers,
                    )
                )
                _data = dict(payload=event.payload)
                current_events_service.create(
                    identity, record.id, _data, event, uow=uow
                )
            record["reviewers"] = new_reviewers


class RequestPayloadComponent(DataComponent):
    """Request variant of DataComponent using dynamic schema."""

    def update(self, identity, data=None, record=None, **kwargs):
        """Update an existing request payload based on permissions."""
        payload = {}
        # take permissions if exist
        permissions = getattr(
            record.type.payload_schema_cls, "field_load_permissions", {}
        )
        if permissions:
            for key in data["payload"]:
                if key in permissions:
                    # permissions should have been checked by now already
                    # so we can assign the new data
                    payload[key] = data["payload"][key]
                else:
                    # keep the old data - no permission to change it
                    # workaround for the lack of patch method
                    payload[key] = record["payload"][key]
            record["payload"] = payload


class RequestLockComponent(ServiceComponent):
    """Component for locking a request."""

    def lock_request(self, identity, record=None, uow=None, **kwargs):
        """Lock a request."""
        event = LogEventType(payload=dict(event="locked"))
        _data = dict(payload=event.payload)
        current_events_service.create(identity, record.id, _data, event, uow=uow)

    def unlock_request(self, identity, record=None, uow=None, **kwargs):
        """Unlock a request."""
        event = LogEventType(payload=dict(event="unlocked"))
        _data = dict(payload=event.payload)
        current_events_service.create(identity, record.id, _data, event, uow=uow)


class RequestCommentFileValidationComponent(ServiceComponent):
    """Component validating files referenced in a request comment."""

    def _check_file_references(
        self, identity, data=None, event=None, uow=None, **kwargs
    ):
        # The new list of files received from the current service call.
        comment_file_ids_next = [
            UUID(file_next["file_id"])
            for file_next in data.get("payload", {}).get("files", [])
        ]

        # Retrieve the existing (persisted) list of files with the given file IDs which are associated to the request.
        request_files_existing = RequestFile.list_by_file_ids(
            event["request_id"], comment_file_ids_next
        )
        request_file_ids_existing = {
            request_file_existing.file.file_id
            for request_file_existing in request_files_existing
        }

        # Check if all the file IDs exist in the database.
        error_messages = []
        for idx, comment_file_id_next in enumerate(comment_file_ids_next):
            if comment_file_id_next not in request_file_ids_existing:
                error_messages.append(
                    {
                        "messages": [_(f"File {str(comment_file_id_next)} not found.")],
                        "field": f"payload.files[{idx}]",
                    }
                )
        if error_messages:
            raise ValidationErrorGroup(error_messages)

    def create(self, identity, data=None, event=None, errors=None, uow=None, **kwargs):
        """Ensures all referenced files exist."""
        self._check_file_references(identity, data=data, event=event, uow=uow, **kwargs)

    def update_comment(
        self, identity, data=None, event=None, request=None, uow=None, **kwargs
    ):
        """Ensures all referenced files exist."""
        self._check_file_references(identity, data=data, event=event, uow=uow, **kwargs)


class RequestCommentFileCleanupComponent(ServiceComponent):
    """Component deleting files which references were removed form a request comment."""

    def update_comment(
        self, identity, data=None, event=None, request=None, uow=None, **kwargs
    ):
        """Delete files not referenced anymore in a comment."""
        # The existing (persisted) list of files associated to the comment.
        file_ids_previous = [
            file_previous["file_id"]
            for file_previous in event["payload"].get("files", [])
        ]

        # The new list of files received from the current service call.
        file_ids_next = [
            file_next["file_id"] for file_next in data["payload"].get("files", [])
        ]

        # Delete files from the persisted list which are not in the new list anymore.
        for file_id_previous in file_ids_previous:
            if file_id_previous not in file_ids_next:
                current_request_files_service.delete_file(
                    identity, event["request_id"], file_id=file_id_previous, uow=uow
                )

    def delete_comment(
        self, identity, data=None, event=None, request=None, uow=None, **kwargs
    ):
        """Delete files not associated to a deleted comment."""
        # The existing (persisted) list of files associated to the comment.
        file_ids_previous = [
            file_previous["file_id"]
            for file_previous in event["payload"].get("files", [])
        ]

        # Delete all the files.
        for file_id_previous in file_ids_previous:
            current_request_files_service.delete_file(
                identity, event["request_id"], file_id=file_id_previous, uow=uow
            )



# -*- coding: utf-8 -*-
#
# Copyright (C) 2025 CERN.
#
# Invenio-Requests is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Validation utilities for request records."""

from ..errors import ChildrenNotSupportedError


def validate_children_allowed(event):
    """Validate that an event type allows children before setting parent_id.

    :param event: The RequestEvent instance to validate.
    :raises ChildrenNotSupportedError: If the event type doesn't support children
                                       but has a parent_id set.
    """
    if event.parent_id is not None:
        # Check if event type supports children
        if hasattr(event, "type") and event.type:
            if not getattr(event.type, "allow_children", False):
                raise ChildrenNotSupportedError(
                    event.type.type_id,
                    f"Event type '{event.type.type_id}' does not support children. "
                    f"Cannot set parent_id for this event type.",
                )



# -*- coding: utf-8 -*-
#
# Copyright (C) 2025 CERN.
#
# Invenio-Requests is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Request Files ui views module."""

from flask import Blueprint, current_app, render_template

from invenio_requests.services.files.errors import RequestFileNotFoundError

from .requests import read_file


def not_found_error(_):
    """Handler for 'Not Found' errors."""
    return render_template(current_app.config["THEME_404_TEMPLATE"]), 404


def create_ui_blueprint(app):
    """Register blueprint routes on app."""
    routes = app.config.get("REQUESTS_ROUTES")

    blueprint = Blueprint(
        "invenio_requests_files",
        __name__,
        template_folder="../templates",
        static_folder="../static",
    )

    # Here we add a non-API endpoint for HTML embedding and downloads,
    # which unlike the API endpoint is not for programmatic file access.
    blueprint.add_url_rule(
        routes["download_file_html"],
        view_func=read_file,
    )

    blueprint.register_error_handler(
        RequestFileNotFoundError,
        not_found_error,
    )

    return blueprint



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021 CERN.
# Copyright (C) 2021 TU Wien.
# Copyright (C) 2021 Northwestern University.
#
# Invenio-Requests is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Resources module."""

from .events import RequestCommentsResource, RequestCommentsResourceConfig
from .files import RequestFilesResource, RequestFilesResourceConfig
from .requests import RequestsResource, RequestsResourceConfig

__all__ = (
    "RequestsResource",
    "RequestsResourceConfig",
    "RequestCommentsResource",
    "RequestCommentsResourceConfig",
    "RequestFilesResource",
    "RequestFilesResourceConfig",
)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2022-2024 CERN.
#
# Invenio-Requests is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Base class for creating custom event types of requests."""

import inspect
from uuid import UUID

import marshmallow as ma
from flask import current_app
from invenio_i18n import lazy_gettext as _
from marshmallow import RAISE, Schema, ValidationError, fields, validate
from marshmallow.validate import OneOf
from marshmallow_utils import fields as utils_fields

from ..proxies import current_requests


def is_uuid(value):
    """Make sure value is a UUID."""
    try:
        UUID(value)
    except (ValueError, TypeError):
        raise ValidationError(
            _("The ID must be an Universally Unique IDentifier (UUID).")
        )


# FIXME: Ideally this should go to records-resources, with an extra argument specifying which config prefix to use.
class RequestsCommentsSanitizedHTML(utils_fields.SanitizedHTML):
    """A subclass of SanitizedHTML that dynamically configures allowed HTML tags and attributes based on application settings."""

    def __init__(self, *args, **kwargs):
        """Initializes RequestsCommentsSanitizedHTML with dynamic tag and attribute settings."""
        super().__init__(tags=None, attrs=None, *args, **kwargs)

    def _deserialize(self, value, attr, data, **kwargs):
        """Deserialize value with dynamic HTML tags and attributes based on Flask app context or defaults."""
        self.tags = (
            current_app.config.get("ALLOWED_HTML_TAGS", [])
            + current_app.config["REQUESTS_COMMENTS_ALLOWED_EXTRA_HTML_TAGS"]
        )
        self.attrs = self.attrs = dict(
            **current_app.config.get("ALLOWED_HTML_ATTRS", {}),
            **current_app.config["REQUESTS_COMMENTS_ALLOWED_EXTRA_HTML_ATTRS"],
        )

        return super()._deserialize(value, attr, data, **kwargs)


class EventType:
    """Base class for event types."""

    type_id = None
    """The unique and constant identifier for this type of event."""

    payload_schema = None
    """Schema for supported payload fields.

    Define it as a dictionary of fields mapping or callable that returns a dictionary:

    .. code-block:: python

        payload_schema = {
            "content": fields.String(),
            # ...
        }

        # or a callable
        def payload_schema():
            return {
                "content": fields.String(),
                # ...
            }
    """

    payload_required = False
    """Require the event payload."""

    allow_children = False
    """Allow this event type to have children (parent-child relationships).

    If True, events of this type can have a parent_id and children.
    If False, attempting to set a parent_id will raise an error.
    """

    def __init__(self, payload=None):
        """Constructor."""
        self.payload = payload or {}

    def __eq__(self, other):
        """Implement comparison test."""
        if inspect.isclass(other):
            # if a class was passed rather than an instance, try to instantiate it
            other = other()
        if isinstance(other, EventType):
            return self.type_id == other.type_id
        elif isinstance(other, str):
            return self.type_id == other
        raise Exception("unknown value")

    def __str__(self):
        """Return str(self)."""
        # Value used by marshmallow schemas to represent the type.
        return self.type_id

    def __repr__(self):
        """Return repr(self)."""
        return f"<RequestEventType '{self.type_id}'>"

    @classmethod
    def _create_marshmallow_schema(cls):
        """Create a marshmallow schema for this request type."""
        # Avoid circular imports
        from invenio_requests.services.schemas import RequestEventSchema

        additional_fields = {}

        # Raise on invalid payload keys
        class PayloadBaseSchema(ma.Schema):
            class Meta:
                unknown = RAISE

        # If a payload schema is defined, add it to the request schema
        if cls.payload_schema is not None:
            payload_schema = cls.payload_schema
            if callable(cls.payload_schema):
                payload_schema = payload_schema()

            payload_required = False
            if cls.payload_required is not None:
                payload_required = cls.payload_required
            additional_fields["payload"] = ma.fields.Nested(
                PayloadBaseSchema.from_dict(payload_schema), required=payload_required
            )

        # Dynamically create a schema from the fields defined
        # by the payload schema dict.
        return RequestEventSchema.from_dict(additional_fields)

    @classmethod
    def marshmallow_schema(cls):
        """Create a schema for the entire request including payload."""
        type_id = cls.type_id
        if type_id not in current_requests._events_schema_cache:
            current_requests._events_schema_cache[type_id] = (
                cls._create_marshmallow_schema()
            )
        return current_requests._events_schema_cache[type_id]


class LogEventType(EventType):
    """Log event type."""

    type_id = "L"

    def payload_schema():
        """Return payload schema as a dictionary."""
        # we need to import here because of circular imports
        from invenio_requests.records.api import RequestEventFormat

        return dict(
            event=fields.String(validate=validate.Length(min=1)),
            content=utils_fields.SanitizedHTML(validate=validate.Length(min=1)),
            format=fields.Str(
                validate=validate.OneOf(choices=[e.value for e in RequestEventFormat]),
                load_default=RequestEventFormat.HTML.value,
            ),
        )


class ReviewersUpdatedType(EventType):
    """Reviewers updated event type."""

    type_id = "R"

    def payload_schema():
        """Return payload schema as a dictionary."""
        # we need to import here because of circular imports
        from invenio_requests.records.api import RequestEventFormat

        return dict(
            event=fields.String(validate=validate.Length(min=1)),
            content=utils_fields.SanitizedHTML(validate=validate.Length(min=1)),
            format=fields.Str(
                validate=validate.OneOf(choices=[e.value for e in RequestEventFormat]),
                load_default=RequestEventFormat.HTML.value,
            ),
            reviewers=fields.List(
                fields.Dict(
                    keys=fields.String(validate=OneOf(("user", "group"))),
                    values=fields.String(required=True),
                )
            ),
        )


class FileDetailsSchema(Schema):
    """File details schema using entity reference format.

    Files are referenced as {"file": "uuid"} to be compatible with
    the entity resolver pattern.
    """

    file_id = fields.String(validate=is_uuid, required=True)


class CommentEventType(EventType):
    """Comment event type."""

    type_id = "C"

    allow_children = True
    """Comments support parent-child relationships (replies to comments)."""

    def payload_schema():
        """Return payload schema as a dictionary."""
        # we need to import here because of circular imports
        from invenio_requests.records.api import RequestEventFormat

        return dict(
            content=RequestsCommentsSanitizedHTML(
                required=True, validate=validate.Length(min=1)
            ),
            format=fields.Str(
                validate=validate.OneOf(choices=[e.value for e in RequestEventFormat]),
                load_default=RequestEventFormat.HTML.value,
            ),
            files=fields.List(fields.Nested(FileDetailsSchema)),
        )

    payload_required = True



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021 TU Wien.
# Copyright (C) 2026 Northwestern University.
#
# Invenio-Requests is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Utility for rendering URI template links."""

import copy

from invenio_records_resources.services import EndpointLink


class RequestEndpointLink(EndpointLink):
    """Shortcut for writing request links."""

    def __init__(self, *args, **kwargs):
        """Constructor."""
        # make sure `params` argument contains "id"
        params = kwargs.get("params", [])
        kwargs["params"] = list(set(params) | {"id"})
        super().__init__(*args, **kwargs)

    @staticmethod
    def vars(record, vars):
        """Update vars used to expand the link."""
        vars.update({"id": record.id})


class RequestTypeDependentEndpointLink(EndpointLink):
    """Class that dynamically delegates to EndpointLink on RequestType's.

    The Requests API (/requests/...) needs to return links that sometimes
    depend on the type of Request. RequestTypeEndpointLink allows that
    by delegating the link to be rendered to one defined in the RequestType
    (where such responsibility should reside) under
    `links_item = {<key>: ...}`.

    RequestTypeDependentEndpointLink guarantees that the EndpointLink
    defined in the RequestType will have "request", "request_type", and
    "request_event" keys in the `vars` dict. "request_event" may be None
    but the key is guaranteed to exist. This provides all the information
    that a RequestType can legitimately require to build EndpointLinks.
    """

    def __init__(
        self,
        key,
        request_retriever=lambda obj, vars: None,
        request_type_retriever=lambda obj, vars: None,
        request_event_retriever=lambda obj, vars: None,
        anchor=lambda obj, vars: None,
    ):
        """Constructor."""
        self._key = key
        self._request_retriever = request_retriever
        self._request_type_retriever = request_type_retriever
        self._request_event_retriever = request_event_retriever
        self._anchor_func = anchor

    def _get_uniform_context(self, obj, context):
        """Fill `context` with retrieved values.

        This is what makes it possible to use
        RequestTypeDependentEndpointLink from a Request- or RequestEvent-
        Service (or any for that matter) as it sets values in keys
        "request", "request_type", "request_event" that an EndpointLink
        defined on a RequestType can rely on.
        """
        ctx = copy.deepcopy(context)
        ctx["request"] = self._request_retriever(obj, ctx)
        ctx["request_type"] = self._request_type_retriever(obj, ctx)
        ctx["request_event"] = self._request_event_retriever(obj, ctx)
        return ctx

    def _retrieve_endpoint_link(self, obj, context):
        """Generate EndpointLink used under the hood.

        Requires _get_uniform_context to have been called to generate
        context.
        """
        # For when an EndpointLink can't be found
        no_op_link = EndpointLink("", when=lambda obj, vars: False)

        # Retrieval
        request_type = context["request_type"]
        if not request_type:
            return no_op_link
        links_item_of_type = getattr(request_type, "links_item", {})
        endpoint_link = links_item_of_type.get(self._key, no_op_link)
        return endpoint_link

    def should_render(self, obj, context):
        """Determine if the link should be rendered."""
        ctx = self._get_uniform_context(obj, context)
        endpoint_link = self._retrieve_endpoint_link(obj, ctx)
        return endpoint_link.should_render(obj, ctx)

    def expand(self, obj, context):
        """Expand/render the endpoint defined on the RequestType."""
        ctx = self._get_uniform_context(obj, context)
        endpoint_link = self._retrieve_endpoint_link(obj, ctx)
        return endpoint_link.expand(obj, ctx)


class RequestListOfCommentsEndpointLink(EndpointLink):
    """Render links for a Request's Comments (Events).

    Note that the RequestCommentsResource uses RequestEventsService.
    """

    def __init__(self, *args, **kwargs):
        """Constructor."""
        params = kwargs.get("params", [])
        kwargs["params"] = list(set(params) | {"request_id"})
        super().__init__(*args, **kwargs)

    @staticmethod
    def vars(record, vars):
        """Update vars used to expand the link."""
        vars.update({"request_id": record.id})


class RequestSingleCommentEndpointLink(EndpointLink):
    """Render links for a Request's Comment (Event)."""

    def __init__(self, *args, **kwargs):
        """Constructor."""
        params = kwargs.get("params", [])
        kwargs["params"] = list(set(params) | {"request_id", "comment_id"})
        super().__init__(*args, **kwargs)

    @staticmethod
    def vars(obj, vars):
        """Update vars used to expand the link."""
        vars.update({"request_id": obj.request_id, "comment_id": obj.id})


class ActionsEndpointLinks:
    """Renders action links.

    This is EndpointLink-input-interface compliant, but renders a dict of
    links. That's why we don't inherit from it directly.
    """

    def __init__(self, endpoint_link):
        """Constructor."""
        self._endpoint_link = endpoint_link

    def should_render(self, obj, context):
        """Always renders to keep with previous/existing interface.

        Previous interface will render `"actions": {}` so we always want to
        at least render {}.
        """
        return True

    def expand(self, obj, context):
        """Expand the endpoints.

        :param obj: Request
        :param context: dict of context data
        """
        links = {}
        request = obj

        for action in request.type.available_actions:
            if action in [request.type.create_action, request.type.delete_action]:
                continue
            ctx = context.copy()
            ctx["action"] = action
            if self._endpoint_link.should_render(request, ctx):
                links[action] = self._endpoint_link.expand(request, ctx)

        return links



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2022-2025 Graz University of Technology.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Celery tasks for requests."""

from datetime import datetime, timezone

from celery import shared_task
from flask import current_app
from invenio_access.permissions import system_identity
from invenio_search.engine import dsl

from invenio_requests.proxies import current_user_moderation_service
from invenio_requests.services.user_moderation.errors import OpenRequestAlreadyExists

from .proxies import current_requests_service


@shared_task
def check_expired_requests():
    """Retrieve expired requests and perform expired action."""
    service = current_requests_service
    now = datetime.now(timezone.utc).isoformat()

    # using scan to get all requests
    requests_list = service.scan(
        identity=system_identity,
        extra_filter=dsl.query.Bool(
            "must",
            must=[
                # somehow querying for '"term", **{"is_expired: True"}' will not return any requests # noqa
                dsl.Q("range", **{"expires_at": {"lte": now}}),
                dsl.Q("term", **{"is_open": True}),
            ],
        ),
    )
    for r in requests_list:
        try:
            service.execute_action(
                identity=system_identity, id_=r["id"], action="expire"
            )
        except Exception as e:
            current_app.logger.warning(e)
            pass


@shared_task(ignore_result=True)
def request_moderation(user_id):
    """Creates a task to request moderation for a user.

    Why this task: when specific actions happen e.g record is published or community
    created, we want to request moderation for a user. However, the moderation service
    might implement heavier operations (e.g. querying) and we don't want to delay the
    publishing of the record.
    """
    try:
        current_user_moderation_service.request_moderation(
            system_identity, user_id=user_id
        )
    except OpenRequestAlreadyExists as ex:
        current_app.logger.warning(ex.description)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021 CERN.
# Copyright (C) 2021 Northwestern University.
# Copyright (C) 2021 TU Wien.
#
# Invenio-Requests is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Services module."""

from .events import RequestEventsService, RequestEventsServiceConfig
from .files import RequestFilesService, RequestFilesServiceConfig
from .requests import RequestsService, RequestsServiceConfig
from .user_moderation import UserModerationRequestService

__all__ = (
    "RequestEventsService",
    "RequestEventsServiceConfig",
    "RequestFilesService",
    "RequestFilesServiceConfig",
    "RequestsService",
    "RequestsServiceConfig",
    "UserModerationRequestService",
)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2025 CERN.
#
# Invenio-Requests is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Request Files views module."""

from flask import g

from invenio_requests.proxies import current_request_files_service


def read_file(pid_value, file_key):
    """Read a file."""
    file = current_request_files_service.read_file(
        identity=g.identity,
        id_=pid_value,
        file_key=file_key,
    )
    return file.send_file(as_attachment=True, restricted=True), 200



# -*- coding: utf-8 -*-
#
# Copyright (C) 2023 CERN.
#
# Invenio-Requests is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.
"""User moderation requests."""

from invenio_i18n import lazy_gettext as _
from invenio_users_resources.proxies import current_users_service
from marshmallow import ValidationError

from invenio_requests.customizations import RequestType, actions


class DeclineUserAction(actions.DeclineAction):
    """Represents a decline action used to block a user."""

    def execute(self, identity, uow):
        """Executes block action."""
        user = self.request.topic.resolve()
        try:
            current_users_service.block(identity, user.id, uow=uow)
        except ValidationError:
            # User already blocked so just ignore
            pass
        super().execute(identity, uow)


class AcceptUserAction(actions.AcceptAction):
    """Represents an accept action used to approve a user."""

    def execute(self, identity, uow):
        """Executes approve action."""
        user = self.request.topic.resolve()
        try:
            current_users_service.approve(identity, user.id, uow=uow)
        except ValidationError:
            # User already approved so just ignore
            pass
        super().execute(identity, uow)


class UserModerationRequest(RequestType):
    """Request to moderate a user."""

    type_id = "user-moderation"
    name = _("User moderation")

    creator_can_be_none = False
    topic_can_be_none = False
    allowed_creator_ref_types = ["group"]
    allowed_receiver_ref_types = ["group"]
    allowed_topic_ref_types = ["user"]

    available_actions = {
        "delete": actions.DeleteAction,
        "submit": actions.SubmitAction,
        "create": actions.CreateAction,
        "cancel": actions.CancelAction,
        # Custom implemented actions
        "accept": AcceptUserAction,
        "decline": DeclineUserAction,
    }



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021 CERN.
# Copyright (C) 2021 Northwestern University.
# Copyright (C) 2022 KTH Royal Institute of Technology
#
# Invenio-Requests is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""RequestEvent Resource Configuration."""

from flask_resources import HTTPJSONException, create_error_handler
from invenio_records_resources.resources import (
    RecordResourceConfig,
    SearchRequestArgsSchema,
)
from marshmallow import fields

from ...errors import (
    ChildrenNotSupportedError,
    NestedChildrenNotAllowedError,
    RequestEventPermissionError,
    RequestLockedError,
)


class RequestCommentsSearchRequestArgsSchema(SearchRequestArgsSchema):
    """Add parameter to parse tags."""

    focus_event_id = fields.UUID()


class RequestCommentsResourceConfig(RecordResourceConfig):
    """Request Events resource configuration."""

    blueprint_name = "request_events"
    url_prefix = "/requests"
    routes = {
        "list": "/<request_id>/comments",
        "item": "/<request_id>/comments/<comment_id>",
        "reply": "/<request_id>/comments/<comment_id>/reply",
        "replies": "/<request_id>/comments/<comment_id>/replies",
        "replies_focused": "/<request_id>/comments/<comment_id>/replies_focused",
        "timeline": "/<request_id>/timeline",
        "timeline_focused": "/<request_id>/timeline_focused",
    }

    # Input
    # WARNING: These "request_*" values have nothing to do with the
    #          "Request" of "RequestEvent". They are related to the Flask
    #          request.
    request_list_view_args = {
        "request_id": fields.UUID(),
        "comment_id": fields.Str(),
    }
    request_item_view_args = {
        "request_id": fields.Str(),
        "comment_id": fields.Str(),
    }

    request_search_args = RequestCommentsSearchRequestArgsSchema

    response_handlers = {
        "application/vnd.inveniordm.v1+json": RecordResourceConfig.response_handlers[
            "application/json"
        ],
        **RecordResourceConfig.response_handlers,
    }

    error_handlers = {
        **RecordResourceConfig.error_handlers,
        RequestLockedError: create_error_handler(
            lambda e: HTTPJSONException(
                code=403,
                description=e.description,
            )
        ),
        RequestEventPermissionError: create_error_handler(
            lambda e: HTTPJSONException(
                code=403,
                description=e.description,
            )
        ),
        NestedChildrenNotAllowedError: create_error_handler(
            lambda e: HTTPJSONException(
                code=400,
                description=str(e),
            )
        ),
        ChildrenNotSupportedError: create_error_handler(
            lambda e: HTTPJSONException(
                code=400,
                description=str(e),
            )
        ),
    }



# -*- coding: utf-8 -*-
#
# Copyright (C) 2025 CERN.
#
# Invenio-Requests is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Requests files errors."""

from invenio_i18n import gettext as _


class RequestFileSizeLimitError(Exception):
    """The provided file size exceeds limit."""

    def __init__(self):
        """Initialise error."""
        super().__init__(_("File size exceeds limit"))


class RequestFileNotFoundError(Exception):
    """The provided file is not found."""

    def __init__(self):
        """Initialise error."""
        super().__init__(_("File not found"))


class RequestFileArgumentMissingError(Exception):
    """The provided file argument is missing."""

    def __init__(self):
        """Initialise error."""
        super().__init__(_("Missing required argument file_key or file_id"))



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021 - 2022 TU Wien.
# Copyright (C) 2021 Northwestern University.
#
# Invenio-Requests is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""API classes for requests in Invenio."""

from enum import Enum
from functools import partial

from flask import current_app
from invenio_db import db
from invenio_files_rest.models import ObjectVersion
from invenio_records.dumpers import SearchDumper
from invenio_records.systemfields import ConstantField, DictField, ModelField
from invenio_records_resources.records.api import FileRecord, Record
from invenio_records_resources.records.systemfields import IndexField

from invenio_requests.records.systemfields.files import RequestFilesField

from ..customizations import RequestState as State
from .dumpers import (
    CalculatedFieldDumperExt,
    GrantTokensDumperExt,
    ParentChildDumperExt,
)
from .models import RequestEventModel, RequestFileMetadata, RequestMetadata
from .systemfields import (
    EntityReferenceField,
    EventTypeField,
    ExpiredStateCalculatedField,
    IdentityField,
    LastActivity,
    LastReply,
    RequestStateCalculatedField,
    RequestStatusField,
    RequestTypeField,
)
from .systemfields.entity_reference import (
    MultiEntityReferenceField,
    check_allowed_creators,
    check_allowed_receivers,
    check_allowed_references,
    check_allowed_reviewers,
    check_allowed_topics,
)


class RequestEventFormat(Enum):
    """Comment/content format enum."""

    HTML = "html"


class RequestEvent(Record):
    """A Request Event."""

    model_cls = RequestEventModel

    dumper = SearchDumper(extensions=[ParentChildDumperExt()])
    """Search dumper with parent-child relationship extension and files extension."""

    # Systemfields
    metadata = None

    schema = ConstantField("$schema", "local://requestevents/requestevent-v1.0.0.json")
    """The JSON Schema to use for validation."""

    request = ModelField(dump=False)
    """The request."""

    request_id = DictField("request_id")
    """The data-layer id of the related Request."""

    type = EventTypeField("type")
    """Request event type system field."""

    index = IndexField(
        "requestevents-requestevent-v1.0.0", search_alias="requestevents"
    )
    """The ES index used."""

    id = ModelField("id")
    """The data-layer id."""

    check_referenced = partial(
        check_allowed_references,
        lambda r: True,  # for system process for now
        lambda r: ["user", "email"],  # only users for now
    )

    created_by = EntityReferenceField("created_by", check_referenced)
    """Who created the event."""

    parent_id = DictField("parent_id")
    """The parent event ID for parent-child relationships."""

    def pre_commit(self):
        """Hook called before committing the record.

        Validates that children are allowed for this event type.
        """
        from .validators import validate_children_allowed

        validate_children_allowed(self)
        super().pre_commit()


def get_files_quota(record=None):
    """Get bucket quota configuration for request files."""
    # Returns quota_size and max_file_size from Flask config
    # with defaults (100MB total quota, 10MB max file size)
    return dict(
        quota_size=current_app.config["REQUESTS_FILES_DEFAULT_QUOTA_SIZE"],
        max_file_size=current_app.config["REQUESTS_FILES_DEFAULT_MAX_FILE_SIZE"],
    )


class RequestFile(FileRecord):
    """Request file API."""

    model_cls = RequestFileMetadata

    @classmethod
    def list_by_file_ids(cls, record_id, file_ids):
        """Get record files by record ID and file IDs."""
        with db.session.no_autoflush:
            query = cls.model_cls.query.join(ObjectVersion).filter(
                cls.model_cls.record_id == record_id,
                ObjectVersion.file_id.in_(file_ids),
                cls.model_cls.is_deleted != True,
            )

            for obj in query:
                yield cls(obj.data, model=obj)


class Request(Record):
    """A generic request record."""

    event_cls = RequestEvent
    """The event class used for request events."""

    model_cls = RequestMetadata
    """The model class for the request."""

    dumper = SearchDumper(
        extensions=[
            CalculatedFieldDumperExt("is_closed"),
            CalculatedFieldDumperExt("is_open"),
            GrantTokensDumperExt("created_by", "receiver", "topic", "reviewers"),
        ]
    )
    """Search dumper with configured extensions."""

    number = IdentityField("number")
    """The request's number (i.e. external identifier)."""

    metadata = None
    """Disabled metadata field from the base class."""

    index = IndexField("requests-request-v1.0.0", search_alias="requests")
    """The Search index to use for the request."""

    schema = ConstantField("$schema", "local://requests/request-v1.0.0.json")
    """The JSON Schema to use for validation."""

    type = RequestTypeField("type")
    """System field for management of the request type.

    This field manages loading of the correct RequestType classes associated with
    `Requests`, based on their `request_type_id` field.
    This is important because the `RequestType` classes are the place where the
    custom request actions are registered.
    """

    topic = EntityReferenceField("topic", check_allowed_topics)
    """Topic (associated object) of the request."""

    created_by = EntityReferenceField("created_by", check_allowed_creators)
    """The entity that created the request."""

    receiver = EntityReferenceField("receiver", check_allowed_receivers)
    """The entity that will receive the request."""

    reviewers = MultiEntityReferenceField("reviewers", check_allowed_reviewers)
    """The entity that will receive the request."""

    status = RequestStatusField("status")
    """The current status of the request."""

    is_closed = RequestStateCalculatedField("status", expected_state=State.CLOSED)
    """Whether or not the current status can be seen as a 'closed' state."""

    is_open = RequestStateCalculatedField("status", expected_state=State.OPEN)
    """Whether or not the current status can be seen as an 'open' state."""

    expires_at = ModelField("expires_at")
    """Expiration date of the request."""

    is_expired = ExpiredStateCalculatedField("expires_at")
    """Whether or not the request is already expired."""

    last_reply = LastReply()
    """The complete last reply event in the request."""

    last_activity_at = LastActivity()
    """The last activity (derived from other fields)."""

    is_locked = DictField("is_locked")
    """Whether or not the request is locked."""

    bucket_id = ModelField(dump=False)
    bucket = ModelField(dump=False)

    # Files NOT dumped or stored in JSON - internal only
    files = RequestFilesField(
        store=False,  # Don't serialize to request JSON
        dump=False,  # Don't include in dumps()
        file_cls=RequestFile,
        delete=False,  # Manual management via service
        create=False,  # Lazy initialization
        bucket_args=get_files_quota,  # Quota config
    )



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021-2025 CERN.
# Copyright (C) 2021-2022 Northwestern University.
# Copyright (C) 2021-2022 TU Wien.
# Copyright (C) 2023 Graz University of Technology.
#
# Invenio-Requests is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""RequestEvents Service."""

import sqlalchemy.exc
from flask import current_app
from flask_principal import AnonymousIdentity
from invenio_i18n import lazy_gettext as _
from invenio_notifications.services.uow import NotificationOp
from invenio_records_resources.services import RecordService, ServiceSchemaWrapper
from invenio_records_resources.services.base.links import LinksTemplate
from invenio_records_resources.services.errors import PermissionDeniedError
from invenio_records_resources.services.records.params import PaginationParam
from invenio_records_resources.services.uow import (
    RecordCommitOp,
    RecordIndexOp,
    unit_of_work,
)
from invenio_search.engine import dsl

from invenio_requests.customizations import CommentEventType
from invenio_requests.customizations.event_types import LogEventType
from invenio_requests.proxies import current_requests_service as requests_service
from invenio_requests.services.results import EntityResolverExpandableField

from ...errors import (
    NestedChildrenNotAllowedError,
    RequestEventPermissionError,
    RequestLockedError,
)
from ...records.api import RequestEventFormat
from ...resolvers.registry import ResolverRegistry


class RequestEventsService(RecordService):
    """Request Events service."""

    def _wrap_schema(self, schema):
        """Wrap schema."""
        return ServiceSchemaWrapper(self, schema)

    @property
    def expandable_fields(self):
        """Get expandable fields for request events.

        Includes:
        - created_by: User or email who created the event

        Note: payload.files: File attachments in comment events are resolved in the result class.
        """
        return [EntityResolverExpandableField("created_by")]

    def links_tpl_factory(self, links, **context):
        """Include context information in the link template.

        This way, the link URLs can be contextualised depending e.g. on the type of the event's
        parent request.
        """
        return LinksTemplate(links, context=context)

    @unit_of_work()
    def create(
        self,
        identity,
        request_id,
        data,
        event_type,
        uow=None,
        expand=False,
        notify=True,
        parent_id=None,
    ):
        """Create a request event (top-level or reply).

        :param request_id: Identifier of the request (data-layer id).
        :param identity: Identity of user creating the event.
        :param dict data: Input data according to the data schema.
        :param event_type: The type of event to create.
        :param parent_id: Optional parent event ID for replies.
        """
        request = self._get_request(request_id)
        self.require_permission(identity, "read", request=request)
        try:
            # If the event is a log, we don't check for permissions to not block logs creation
            if event_type.type_id != LogEventType.type_id:
                # Check permission based on whether this is a reply or top-level comment
                permission = "reply_comment" if parent_id else "create_comment"
                self.require_permission(identity, permission, request=request)
        except PermissionDeniedError:
            if current_app.config.get(
                "REQUESTS_LOCKING_ENABLED", False
            ) and request.get("is_locked", False):
                raise RequestLockedError(
                    description=_("Commenting is now locked for this conversation.")
                )
            else:
                raise RequestEventPermissionError(
                    description=_(
                        "You do not have permission to comment on this conversation."
                    )
                )

        # Validate that nested children (reply to reply) are not allowed
        if parent_id is not None:
            parent_event = self._get_event(parent_id)
            if parent_event.parent_id is not None:
                raise NestedChildrenNotAllowedError()

        # Validate data (if there are errors, .load() raises)
        schema = self._wrap_schema(event_type.marshmallow_schema())

        data, errors = schema.load(
            data,
            context={"identity": identity},
        )

        event = self.record_cls.create(
            {},
            request=request.model,
            request_id=str(request.id),
            type=event_type,
        )
        event.update(data)
        event.created_by = self._get_creator(identity, request=request)

        # Set parent_id for replies
        if parent_id is not None:
            event.parent_id = parent_id

        # Run components
        self.run_components(
            "create",
            identity,
            data=data,
            event=event,
            errors=errors,
            uow=uow,
        )

        # Persist record (DB and index)
        uow.register(RecordCommitOp(event, indexer=self.indexer))

        # Reindex the request to update events-related computed fields
        # NOTE: We're not reindexing if the event is a deletion log event, since the
        # associated request will anyways be deleted from the index. Ideally, instead of
        # checking this here, the Unit of Work, should be able to optimize away the
        # unecessary reindexing operations (i.e. if there's a RecordDeleteOp registered,
        # all previous RecordIndexOps for the same record should be ignored).
        is_delete_event = (
            isinstance(event_type, LogEventType) or event_type is LogEventType
        ) and (data.get("payload", {}).get("event") == "deleted")
        if not is_delete_event:
            uow.register(RecordIndexOp(request, indexer=requests_service.indexer))

        if notify and event_type is CommentEventType:
            # Use different notification builder for replies vs top-level comments
            if parent_id:
                builder = request.type.reply_notification_builder
            else:
                builder = request.type.comment_notification_builder

            uow.register(NotificationOp(builder.build(request, event)))

        return self.result_item(
            self,
            identity,
            event,
            schema=schema,
            links_tpl=self.links_tpl_factory(
                self.config.links_item, request=request, request_type=request.type
            ),
            expandable_fields=self.expandable_fields,
            expand=expand,
            request=request,
        )

    def read(self, identity, id_, expand=False):
        """Retrieve a record."""
        event = self._get_event(id_)
        request = self._get_request(event.request_id)

        self.require_permission(identity, "read", request=request)

        return self.result_item(
            self,
            identity,
            event,
            schema=self._wrap_schema(event.type.marshmallow_schema()),
            links_tpl=self.links_tpl_factory(
                self.config.links_item, request=request, request_type=request.type
            ),
            expandable_fields=self.expandable_fields,
            expand=expand,
            request=request,
        )

    @unit_of_work()
    def update(self, identity, id_, data, revision_id=None, uow=None, expand=False):
        """Update a comment (only comments can be updated)."""
        event = self._get_event(id_)
        request = self._get_request(event.request.id)
        try:
            self.require_permission(
                identity, "update_comment", request=request, event=event
            )
        except PermissionDeniedError:
            if current_app.config.get(
                "REQUESTS_LOCKING_ENABLED", False
            ) and request.get("is_locked", False):
                raise RequestLockedError(
                    description=_("Updating is now locked for this comment.")
                )
            else:
                raise RequestEventPermissionError(
                    description=_("You do not have permission to update this comment.")
                )
        self.check_revision_id(event, revision_id)

        if event.type != CommentEventType:
            raise RequestEventPermissionError(
                description=_("You cannot update this event.")
            )

        schema = self._wrap_schema(event.type.marshmallow_schema())
        data, errors = schema.load(
            data,
            context=dict(
                identity=identity, record=event, request=request, event_type=event.type
            ),
        )

        event["payload"]["content"] = data["payload"]["content"]
        event["payload"]["format"] = data["payload"]["format"]

        # Run components
        self.run_components(
            "update_comment",
            identity,
            data=data,
            event=event,
            request=request,
            uow=uow,
        )

        # Updating the list of files after running components,
        # so that RequestCommentFileCleanupComponent can compare the list of files.
        if "files" in data["payload"]:
            event["payload"]["files"] = data["payload"]["files"]

        # Persist record (DB and index)
        uow.register(RecordCommitOp(event, indexer=self.indexer))

        # Reindex the request to update events-related computed fields
        uow.register(RecordIndexOp(request, indexer=requests_service.indexer))

        return self.result_item(
            self,
            identity,
            event,
            schema=schema,
            links_tpl=self.links_tpl_factory(
                self.config.links_item, request=request, request_type=request.type
            ),
            expandable_fields=self.expandable_fields,
            expand=expand,
            request=request,
        )

    @unit_of_work()
    def delete(self, identity, id_, revision_id=None, uow=None):
        """Delete a comment (only comments can be deleted)."""
        event = self._get_event(id_)
        request_id = event.request_id
        request = self._get_request(request_id)

        # Permissions
        self.require_permission(
            identity, "delete_comment", request=request, event=event
        )
        self.check_revision_id(event, revision_id)

        if event.type != CommentEventType:
            raise RequestEventPermissionError(
                description=_("You cannot delete this event.")
            )

        # update the event for the deleted comment with a LogEvent
        event.type = LogEventType
        schema = self._wrap_schema(event.type.marshmallow_schema())
        data = dict(
            payload=dict(
                event="comment_deleted",
                content="comment was deleted",
                format=RequestEventFormat.HTML.value,
            )
        )
        data, _errors = schema.load(
            data,
            context=dict(identity=identity, record=event, event_type=event.type),
        )

        for data_key, data_value in data["payload"].items():
            if data_key != "files":
                event["payload"][data_key] = data_value

        # Run components
        self.run_components(
            "delete_comment",
            identity,
            data=data,
            event=event,
            request=request,
            uow=uow,
        )

        # Updating the list of files after running components,
        # so that RequestCommentFileCleanupComponent can get the list of files.
        if "files" in data["payload"]:
            event["payload"]["files"] = data["payload"]["files"]

        # Commit the updated comment
        uow.register(RecordCommitOp(event, indexer=self.indexer))

        # Reindex the request to update events-related computed fields
        uow.register(RecordIndexOp(request, indexer=requests_service.indexer))

        return True

    def search(
        self,
        identity,
        request_id,
        params=None,
        search_preference=None,
        preview_size=None,
        **kwargs,
    ):
        """Search for events (timeline) for a given request.

        Returns all top-level events (parent comments without parent_id) for the request.
        For parents that have children, includes a preview via inner_hits using
        OpenSearch join relationships.

        """
        params = params or {}
        params.setdefault("sort", "oldest")
        expand = kwargs.pop("expand", False)

        # Permissions - guarded by the request's can_read.
        request = self._get_request(request_id)
        self.require_permission(identity, "read", request=request)

        # Build query for top-level events (parents) with optional children preview
        # Uses join relationships to include children via inner_hits when they exist
        search = self._search(
            "search",
            identity,
            params,
            search_preference,
            permission_action="unused",
            **kwargs,
        )

        # Query structure:
        # - must: filter to this request
        # - must_not: exclude children (events with parent_id)
        # - should: optionally add children preview via has_child + inner_hits
        search = search.query(
            "bool",
            must=[
                dsl.Q("term", request_id=str(request.id)),
            ],
            must_not=[
                dsl.Q("exists", field="parent_id"),  # Exclude replies
            ],
            should=[self._timeline_query_child_preview(preview_size)],
            minimum_should_match=0,  # Make should clause optional to return parents without children
        )

        # Execute search
        search_result = search.execute()

        return self.result_list(
            self,
            identity,
            search_result,
            params,
            links_tpl=self.links_tpl_factory(
                self.config.links_search, request_id=str(request.id), args=params
            ),
            links_item_tpl=self.links_tpl_factory(
                self.config.links_item, request=request, request_type=request.type
            ),
            expandable_fields=self.expandable_fields,
            expand=expand,
            request=request,
        )

    def focused_list(
        self,
        identity,
        request_id,
        focus_event_id,
        page_size,
        expand=False,
        search_preference=None,
        preview_size=None,
    ):
        """Return a page of results focused on a given event, or the first page if the event is not found.

        Only searches parent comments (excludes child comments/replies).
        If the focused event is a reply (child comment), the page containing its parent will be returned.
        """
        # Permissions - guarded by the request's can_read.
        request = self._get_request(request_id)
        self.require_permission(identity, "read", request=request)

        # If a specific event ID is requested, we need to work out the corresponding page number.
        focus_event = None
        try:
            focus_event = self._get_event(focus_event_id)
            # Make sure the event belongs to the request, otherwise the `require_permission` call above
            # might not be valid for this particular event.
            if str(focus_event.request_id) != str(request_id):
                raise PermissionDeniedError()

            if focus_event.parent_id is not None:
                focus_event = self._get_event(focus_event.parent_id)
        except sqlalchemy.exc.NoResultFound:
            # Silently ignore
            pass

        params = {"sort": "oldest", "size": page_size}

        # Build filter to only include parent comments (exclude child comments)
        parent_filter = dsl.Q(
            "bool",
            must=[dsl.Q("term", request_id=str(request.id))],
            must_not=[dsl.Q("exists", field="parent_id")],  # Exclude replies
            should=[self._timeline_query_child_preview(preview_size)],
            minimum_should_match=0,
        )
        search = self._search(
            "search",
            identity,
            params,
            search_preference,
            permission_action="unused",
            extra_filter=parent_filter,
            versioning=False,
        )

        page = 1
        if focus_event is not None:
            num_older_than_event = search.filter(
                "range", created={"lt": focus_event.created}
            ).count()
            page = num_older_than_event // page_size + 1

        # Re run the pagination param interpreter to update the search with the new page number
        params.update(page=page)
        search = PaginationParam(self.config.search).apply(identity, search, params)

        # We deactivated versioning before (it doesn't apply for count queries) so we need to re-enable it.
        search_result = search.params(version=True).execute()
        return self.result_list(
            self,
            identity,
            search_result,
            params,
            links_tpl=self.links_tpl_factory(
                self.config.links_search, request_id=str(request.id), args=params
            ),
            links_item_tpl=self.links_tpl_factory(
                self.config.links_item, request=request, request_type=request.type
            ),
            expandable_fields=self.expandable_fields,
            expand=expand,
            request=request,
        )

    def scan(
        self,
        identity,
        request_id,
        params=None,
        search_preference=None,
        expand=False,
        extra_filter=None,
        **kwargs,
    ):
        """Scan for events matching the querystring."""
        request = self._get_request(request_id)
        self.require_permission(identity, "read", request=request)

        # Prepare and execute the search as scan()
        params = params or {}
        search_result = self._search(
            "scan",
            identity,
            params,
            search_preference,
            extra_filter=extra_filter,
            **kwargs,
        ).scan()

        return self.result_list(
            self,
            identity,
            search_result,
            params,
            links_tpl=self.links_tpl_factory(
                self.config.links_search, request_id=str(request.id), args=params
            ),
            links_item_tpl=self.links_tpl_factory(
                self.config.links_item, request=request, request_type=request.type
            ),
            expandable_fields=self.expandable_fields,
            expand=expand,
            request=request,
        )

    def get_comment_replies(
        self, identity, parent_id, params=None, search_preference=None, **kwargs
    ):
        """Get paginated replies for a specific comment.

        :param identity: Identity of user.
        :param parent_id: ID of the parent comment.
        :param params: Query parameters (page, size, sort, etc.).
        :param search_preference: Search preference.
        :returns: Paginated list of reply events.
        """
        params = params or {}
        params.setdefault("sort", "oldest")

        expand = kwargs.pop("expand", False)

        # Get the parent event to verify permissions and get request_id
        parent_event = self._get_event(parent_id)
        request = self._get_request(parent_event.request_id)

        # Permissions - guarded by the request's can_read
        self.require_permission(identity, "read", request=request)

        # Prepare and execute the search for children
        replies_filter = dsl.Q(
            "bool",
            must=[
                dsl.Q("term", request_id=str(request.id)),
                dsl.Q("term", parent_id=parent_id),
            ],
        )
        search = self._search(
            "search",
            identity,
            params,
            search_preference,
            permission_action="unused",
            extra_filter=replies_filter,
            **kwargs,
        )

        search_result = search.execute()

        return self.result_list(
            self,
            identity,
            search_result,
            params,
            links_tpl=self.links_tpl_factory(
                self.config.links_replies,
                request_id=str(request.id),
                comment_id=parent_id,
                args=params,
            ),
            links_item_tpl=self.links_tpl_factory(
                self.config.links_item, request=request, request_type=request.type
            ),
            expandable_fields=self.expandable_fields,
            expand=expand,
            request=request,
        )

    def focused_reply_list(
        self,
        identity,
        parent_id,
        focus_reply_event_id,
        page_size,
        expand=False,
        search_preference=None,
        params=None,
    ):
        """Return a page of results focused on a given reply event, or the first page if the event is not found.

        Only searches reply comments (excludes parents comments/replies).
        """
        parent_event = self._get_event(parent_id)
        request = self._get_request(parent_event.request_id)
        # Permissions - guarded by the request's can_read.
        self.require_permission(identity, "read", request=request)

        # If a specific event ID is requested, we need to work out the corresponding page number.
        focus_event = None
        try:
            focus_event = self._get_event(focus_reply_event_id)
            # Make sure the event belongs to the request, otherwise the `require_permission` call above
            # might not be valid for this particular event.
            if str(focus_event.request_id) != str(parent_event.request_id):
                raise PermissionDeniedError()

            if str(focus_event.parent_id) != str(parent_id):
                raise Exception("Focus event must be a reply to the parent.")

            if focus_event.parent_id is None:
                raise Exception("Cannot focus on non-reply event.")
        except sqlalchemy.exc.NoResultFound:
            # Silently ignore
            pass

        params = params or {}
        params.setdefault("sort", "newest")
        params.setdefault("size", page_size)

        replies_filter = dsl.Q(
            "bool",
            must=[
                dsl.Q("term", request_id=str(request.id)),
                dsl.Q("term", parent_id=parent_id),
            ],
        )
        search = self._search(
            "search",
            identity,
            params,
            search_preference,
            permission_action="unused",
            extra_filter=replies_filter,
            versioning=False,
        )

        page = 1
        if focus_event is not None:
            num_newer_than_event = search.filter(
                "range", created={"gt": focus_event.created}
            ).count()
            page = num_newer_than_event // page_size + 1

        # Re run the pagination param interpreter to update the search with the new page number
        params.update(page=page)
        search = PaginationParam(self.config.search).apply(identity, search, params)

        # We deactivated versioning before (it doesn't apply for count queries) so we need to re-enable it.
        search_result = search.params(version=True).execute()
        return self.result_list(
            self,
            identity,
            search_result,
            params,
            links_tpl=self.links_tpl_factory(
                self.config.links_replies,
                request_id=str(request.id),
                comment_id=parent_id,
                args=params,
            ),
            links_item_tpl=self.links_tpl_factory(
                self.config.links_item, request=request, request_type=request.type
            ),
            expandable_fields=self.expandable_fields,
            expand=expand,
            request=request,
        )

    # Utilities
    @property
    def request_cls(self):
        """Get associated request class."""
        return self.config.request_cls

    def _get_request(self, request_id):
        """Get associated request."""
        # If it's already a request, return it
        if isinstance(request_id, self.request_cls):
            return request_id
        return self.request_cls.get_record(request_id)

    def _get_event(self, event_id, with_deleted=True):
        """Get associated event_id."""
        return self.record_cls.get_record(event_id, with_deleted=with_deleted)

    def _get_creator(self, identity, request=None):
        """Get the creator dict from the identity."""
        creator = None
        if isinstance(identity, AnonymousIdentity):
            # not ideal - assumes that comment is created by same person
            # who created a request - solution for guest users
            creator = request["created_by"]

        referenced_creator = (
            ResolverRegistry.reference_entity(creator, raise_=True)
            if creator is not None
            else ResolverRegistry.reference_identity(identity)
        )
        return referenced_creator

    def _timeline_query_child_preview(self, preview_size):
        """Return an OpenSearch query to include a size-limited preview of replies to a parent comment."""
        if preview_size is None:
            preview_size = current_app.config["REQUESTS_COMMENT_PREVIEW_LIMIT"]
        return dsl.Q(
            "has_child",
            type="child",
            query=dsl.Q("match_all"),
            score_mode="none",
            inner_hits={
                "name": "replies_preview",
                "size": preview_size,
                "sort": [{"created": "desc"}],
            },
        )



# -*- coding: utf-8 -*-
#
# Copyright (C) 2025 CERN.
#
# Invenio-Requests is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Request Files Resource module."""

from .config import RequestFilesResourceConfig
from .resource import RequestFilesResource

__all__ = (
    "RequestFilesResource",
    "RequestFilesResourceConfig",
)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021 CERN.
# Copyright (C) 2021-2025 Northwestern University.
# Copyright (C) 2021 TU Wien.
#
# Invenio-Requests is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Request Events Service Config."""

from uuid import UUID

from invenio_indexer.api import RecordIndexer
from invenio_records_resources.services import (
    RecordServiceConfig,
    ServiceSchemaWrapper,
    pagination_endpoint_links,
)
from invenio_records_resources.services.base.config import ConfiguratorMixin, FromConfig
from invenio_records_resources.services.records.results import (
    RecordItem,
    RecordList,
)

from ...proxies import (
    current_request_files_service,
)
from ...records.api import Request, RequestEvent, RequestFile
from ...services.links import (
    RequestSingleCommentEndpointLink,
    RequestTypeDependentEndpointLink,
)
from ..permissions import PermissionPolicy
from ..schemas import RequestEventSchema


def _expand_files_for_projections(projections, request, identity):
    """Expand file details for multiple projections in a single batch query.

    :param projections: List of projection dictionaries
    :param request: The request object for context
    :param identity: The identity for link generation
    """
    if not projections or not request:
        return

    request_id = request.id

    # Collect all file IDs from all projections
    all_file_ids = set()
    for projection in projections:
        payload_files = projection.get("payload", {}).get("files", [])
        for file_obj in payload_files:
            if "file_id" in file_obj:
                all_file_ids.add(UUID(file_obj["file_id"]))

    # Early return if no files to expand
    if not all_file_ids:
        return

    # Single batch query to fetch all files
    request_files = RequestFile.list_by_file_ids(request_id, list(all_file_ids))
    request_files_by_file_id = {
        request_file.file.file_id: request_file for request_file in request_files
    }

    # Create link template using the files service factory
    file_links_tpl = current_request_files_service.links_tpl_factory(
        current_request_files_service.config.links_item,
        request_type=str(request.type),
    )

    # Expand files in each projection
    for projection in projections:
        payload_files = projection.get("payload", {}).get("files", [])

        # Nothing to expand if no files
        if not payload_files:
            continue

        # Build expanded files list
        expanded_files = []
        for payload_file in payload_files:
            if "file_id" not in payload_file:
                continue

            payload_file_uuid = UUID(payload_file["file_id"])
            file_details = request_files_by_file_id.get(payload_file_uuid)

            # Only expand if the file is found in the database
            if file_details:
                expanded_files.append(
                    {
                        "file_id": str(file_details.file.file_id),
                        "key": file_details.file.key,
                        "original_filename": file_details.model.data[
                            "original_filename"
                        ],
                        "size": file_details.file.size,
                        "mimetype": file_details.file.mimetype,
                        "created": file_details.file.created,
                        # Use link template to expand links
                        "links": file_links_tpl.expand(identity, file_details),
                    }
                )

        # Add expanded files to the projection's expanded field
        if expanded_files:
            if "expanded" not in projection:
                projection["expanded"] = {}
            projection["expanded"]["files"] = expanded_files


class RequestEventItem(RecordItem):
    """RequestEvent result item."""

    def __init__(self, *args, **kwargs):
        """Constructor."""
        request = kwargs.pop("request", None)
        super().__init__(*args, **kwargs)
        self._request = request

    @property
    def id(self):
        """Id property."""
        return self._record.id

    @property
    def data(self):
        """Property to get the record projection with request in context."""
        if self._data:
            return self._data

        self._data = self._schema.dump(
            self._obj,
            context=dict(
                identity=self._identity,
                record=self._record,
                request=self._request,  # Need to pass the request to the schema to get the permissions to check if locked
            ),
        )
        if self._links_tpl:
            self._data["links"] = self.links

        if self._nested_links_item:
            for link in self._nested_links_item:
                link.expand(self._identity, self._record, self._data)

        if self._expand and self._fields_resolver:
            self._fields_resolver.resolve(self._identity, [self._data])
            fields = self._fields_resolver.expand(self._identity, self._data)
            self._data["expanded"] = fields

            # Expand file details if present
            _expand_files_for_projections([self._data], self._request, self._identity)

        return self._data


class RequestEventList(RecordList):
    """RequestEvent result item."""

    def __init__(self, *args, **kwargs):
        """Constructor."""
        request = kwargs.pop("request", None)
        super().__init__(*args, **kwargs)
        self._request = request

    def to_dict(self):
        """Return result as a dictionary with expanded fields for parents and children."""
        # Call parent to handle standard expansion
        res = super().to_dict()

        # Additionally expand children fields if present
        if self._expand and self._fields_resolver:
            self._expand_children_fields(res["hits"]["hits"])

            # Batch expand file details for all hits
            self._batch_expand_file_details(res["hits"]["hits"])

        return res

    def _expand_children_fields(self, hits):
        """Apply field expansion to children arrays in hits.

        :param hits: List of hit dictionaries that may contain children arrays
        """
        # Collect all children from all hits
        all_children = []
        for hit in hits:
            if "children" in hit and hit["children"]:
                all_children.extend(hit["children"])

        if all_children:
            # Batch resolve all children at once for efficiency
            self._fields_resolver.resolve(self._identity, all_children)

            # Expand each child individually
            for child in all_children:
                fields = self._fields_resolver.expand(self._identity, child)
                child["expanded"] = fields

    def _batch_expand_file_details(self, hits):
        """Batch expand file details for all hits in a single database query.

        :param hits: List of hit dictionaries that may contain payload.files
        """
        # Collect all projections (parents and children)
        all_projections = []
        for hit in hits:
            all_projections.append(hit)
            # Also include children
            all_projections.extend(hit.get("children", []))

        # Use common expansion function
        _expand_files_for_projections(all_projections, self._request, self._identity)

    @property
    def hits(self):
        """Iterator over the hits."""
        for hit in self._results:
            # Load dump
            record = self._service.record_cls.loads(hit.to_dict())

            # Project the record
            schema = ServiceSchemaWrapper(
                self._service, record.type.marshmallow_schema()
            )
            projection = schema.dump(
                record,
                context=dict(
                    identity=self._identity,
                    record=record,
                    request=self._request,  # Need to pass the request to the schema to get the permissions to check if locked
                    meta=hit.meta,
                ),
            )

            # Handle inner_hits from has_child queries (join relationship approach)
            # Initialize defaults for parents without children
            projection["children"] = []
            projection["children_count"] = 0

            if (
                hasattr(hit.meta, "inner_hits")
                and "replies_preview" in hit.meta.inner_hits
            ):
                # Extract children from inner_hits
                inner_hits_data = hit.meta.inner_hits.replies_preview.hits
                inner_children = inner_hits_data.hits
                total_children = inner_hits_data.total.value

                projection["children_count"] = total_children

                for inner_hit in inner_children:
                    # Load child record
                    child_record = self._service.record_cls.loads(
                        inner_hit["_source"].to_dict()
                    )

                    # Project child record
                    child_schema = ServiceSchemaWrapper(
                        self._service, child_record.type.marshmallow_schema()
                    )
                    child_projection = child_schema.dump(
                        child_record,
                        context=dict(
                            identity=self._identity,
                            record=child_record,
                            request=self._request,  # Need to pass the request to the schema to get the permissions to check if locked
                            meta=hit.meta,
                        ),
                    )

                    if self._links_item_tpl:
                        child_projection["links"] = self._links_item_tpl.expand(
                            self._identity, child_record
                        )

                    projection["children"].append(child_projection)

            if self._links_item_tpl:
                projection["links"] = self._links_item_tpl.expand(
                    self._identity, record
                )

            yield projection


class ParentChildRecordIndexer(RecordIndexer):
    """Parent-Child Record Indexer placeholder."""

    def _prepare_record(self, record, index, arguments=None, **kwargs):
        """Prepare request-event data for indexing.

        Pass routing information for parent-child relationships.
        """
        data = super()._prepare_record(record, index, arguments, **kwargs)
        if hasattr(record, "parent_id") and record.parent_id:
            arguments["routing"] = str(record.parent_id)
        return data


def request_event_anchor(_, vars):
    """Generate the anchor for request events.

    This is also called for requests, so we need to return None if it isn't being called for a request event.
    """
    request_event = vars["request_event"]
    if request_event is None:
        return None

    if request_event.parent_id is not None:
        return f"commentevent-{request_event.parent_id}_{request_event.id}"
    else:
        return f"commentevent-{request_event.id}"


class RequestEventsServiceConfig(RecordServiceConfig, ConfiguratorMixin):
    """Config."""

    service_id = "request_events"

    request_cls = Request
    permission_policy_cls = FromConfig(
        "REQUESTS_PERMISSION_POLICY", default=PermissionPolicy
    )
    schema = RequestEventSchema
    record_cls = RequestEvent
    result_item_cls = RequestEventItem
    result_list_cls = RequestEventList
    indexer_queue_name = "events"
    indexer_cls = ParentChildRecordIndexer

    # ResultItem configurations
    links_item = {
        # Note that `request_events` is the name of the blueprint for
        # the RequestCommentsResource actually.
        "self": RequestSingleCommentEndpointLink("request_events.read"),
        # Keeps assumption that there is no dedicated UI endpoint for
        # a RequestEvent i.e., RequestType is what determines the UI endpoint
        "self_html": RequestTypeDependentEndpointLink(
            key="self_html",
            request_retriever=lambda obj, vars: vars.get("request"),
            request_type_retriever=lambda obj, vars: vars.get("request_type"),
            # The presence of request_event_retriever
            # provides for further differentiation
            request_event_retriever=lambda obj, vars: obj,
            anchor=request_event_anchor,
        ),
        "reply": RequestSingleCommentEndpointLink(
            "request_events.reply",
            # The reply link is only shown if the request_event is top-level:
            # to send stronger signal to client that only top-level comments
            # can be replied to + no need to parse link to figure if parent or
            # current comment is targeted
            when=lambda obj, vars: obj.parent_id is None,
        ),
        "replies": RequestSingleCommentEndpointLink(
            "request_events.get_replies",
            # The replies link is only shown if the request_event is top-level
            # only case where there *can* be replies
            when=lambda obj, vars: obj.parent_id is None,
        ),
        "replies_focused": RequestSingleCommentEndpointLink(
            "request_events.focused_replies",
            when=lambda obj, vars: obj.parent_id is None,
        ),
    }

    links_search = pagination_endpoint_links(
        "request_events.search",
        params=["request_id"],
    )

    links_replies = pagination_endpoint_links(
        "request_events.get_replies",
        params=["request_id", "comment_id"],
    )

    components = FromConfig(
        "REQUESTS_EVENTS_SERVICE_COMPONENTS",
        default=[],
    )



# -*- coding: utf-8 -*-
#
# Copyright (C) 2025 CERN.
#
# Invenio-Requests is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Request files service."""

from os.path import splitext
from uuid import UUID

from base32_lib import base32
from flask import current_app
from invenio_records_resources.services import FileService, ServiceSchemaWrapper
from invenio_records_resources.services.base.links import LinksTemplate
from invenio_records_resources.services.uow import RecordCommitOp, unit_of_work

from invenio_requests.services.files.errors import (
    RequestFileArgumentMissingError,
    RequestFileNotFoundError,
    RequestFileSizeLimitError,
)


class RequestFilesService(FileService):
    """Service for managing request file attachments."""

    def _wrap_schema(self, schema):
        """Wrap schema."""
        return ServiceSchemaWrapper(self, schema)

    # Utilities
    @property
    def request_cls(self):
        """Get associated request class."""
        return self.config.request_cls

    def links_tpl_factory(self, links, **context):
        """Include context information in the link template.

        This way, the link URLs can be contextualised depending e.g. on the type of the event's
        parent request.
        """
        return LinksTemplate(links, context=context)

    @unit_of_work()
    def create_file(self, identity, id_, key, stream, content_length, uow=None):
        """Upload a file in a single operation (simple endpoint).

        Convenience method that combines init/upload/commit into one operation.
        """
        # Resolve and check permissions
        request = self.request_cls.get_record(id_)
        self.require_permission(identity, "manage_files", request=request)

        # File size validation
        max_file_size = current_app.config["REQUESTS_FILES_DEFAULT_MAX_FILE_SIZE"]
        if content_length > max_file_size:
            raise RequestFileSizeLimitError()

        # Unique key generation
        unique_id = base32.generate(length=10, split_every=5, checksum=True)
        key_root, key_ext = splitext(key)
        unique_key = f"{key_root}-{unique_id}{key_ext}"

        # Lazy files enable initialization (for requests created before request files support)
        if request.files is None:
            # RequestFilesField allows to set the files field after initialization.
            request.files = {"enabled": True}

        # Lazy bucket initialization
        if request.files.bucket is None:
            request.files.create_bucket()

        request.files[unique_key] = stream
        # Store the original filename in RequestFileMetadata.json (field from RecordMetadataBase)
        request.files[unique_key].model.data = {"original_filename": key}

        uow.register(RecordCommitOp(request))

        return self.result_item(
            self,
            identity,
            request.files[unique_key],
            schema=self._wrap_schema(self.config.schema),
            links_tpl=self.links_tpl_factory(
                self.config.links_item, request_type=str(request.type)
            ),
        )

    @unit_of_work()
    def delete_file(self, identity, id_, file_key=None, file_id=None, uow=None):
        """Delete a specific file."""
        # Called explicitly via API or by frontend when file removed from comment

        if file_key is None and file_id is None:
            raise RequestFileArgumentMissingError()

        # Resolve and check permissions
        request = self.request_cls.get_record(id_)
        self.require_permission(identity, "manage_files", request=request)

        if file_key is not None:
            deleted_file = request.files.pop(file_key, None)
        else:
            deleted_file = None
            file_id_uuid = UUID(file_id)
            for request_file_key in request.files:
                request_file = request.files[request_file_key]
                if request_file.file.id == file_id_uuid:
                    deleted_file = request.files.pop(request_file_key, None)
                    break

        if deleted_file is None:
            raise RequestFileNotFoundError()

        deleted_file.delete(force=True)

        uow.register(RecordCommitOp(request))

        return self.result_item(
            self,
            identity,
            deleted_file,
            schema=self._wrap_schema(self.config.schema),
            links_tpl=self.links_tpl_factory(
                self.config.links_item, request_type=str(request.type)
            ),
        )

    def read_file(self, identity, id_, file_key):
        """Retrieve file content for download/display."""
        # Resolve and check permissions
        request = self.request_cls.get_record(id_)
        self.require_permission(identity, "read_files", request=request)

        # Return file stream
        request_file = request.files.get(file_key)
        if request_file is None:
            raise RequestFileNotFoundError()

        return self.file_result_item(
            self,
            identity,
            request_file,
            request,
            links_tpl=self.file_links_item_tpl(id_),
        )



#
# This file is part of Invenio.
# Copyright (C) 2025 CERN.
# Copyright (C) 2026 Graz University of Technology.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Alter datetime columns to utc aware datetime columns."""

from invenio_db.utils import (
    update_table_columns_column_type_to_datetime,
    update_table_columns_column_type_to_utc_datetime,
)

# revision identifiers, used by Alembic.
revision = "74b23178bfbe"
down_revision = "1763728177"
branch_labels = ()
depends_on = None


def upgrade():
    """Upgrade database."""
    for table_name in ["request_metadata", "request_events"]:
        update_table_columns_column_type_to_utc_datetime(table_name, "created")
        update_table_columns_column_type_to_utc_datetime(table_name, "updated")
    update_table_columns_column_type_to_utc_datetime("request_metadata", "expires_at")


def downgrade():
    """Downgrade database."""
    for table_name in ["request_metadata", "request_events"]:
        update_table_columns_column_type_to_datetime(table_name, "created")
        update_table_columns_column_type_to_datetime(table_name, "updated")
    update_table_columns_column_type_to_datetime("request_metadata", "expires_at")



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021 CERN.
# Copyright (C) 2021 TU Wien.
#
# Invenio-Requests is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""View functions for the requests."""


def create_requests_bp(app):
    """Create requests blueprint."""
    ext = app.extensions["invenio-requests"]
    return ext.requests_resource.as_blueprint()


def create_request_events_bp(app):
    """Create request events blueprint."""
    ext = app.extensions["invenio-requests"]
    # `request_events_resource` is really the RequestCommentsResource
    return ext.request_events_resource.as_blueprint()


def create_request_files_bp(app):
    """Create request files blueprint."""
    ext = app.extensions["invenio-requests"]
    return ext.request_files_resource.as_blueprint()
